@@ -1,7 +1,7 @@
-/*	$Id: man_term.c,v 1.121 2011/09/21 09:57:13 schwarze Exp $ */
+/*	$Id: man_term.c,v 1.149 2014/06/20 23:02:31 schwarze Exp $ */
 /*
- * Copyright (c) 2008, 2009, 2010, 2011 Kristaps Dzonsons <kristaps@bsd.lv>
- * Copyright (c) 2010, 2011 Ingo Schwarze <schwarze@openbsd.org>
+ * Copyright (c) 2008-2012 Kristaps Dzonsons <kristaps@bsd.lv>
+ * Copyright (c) 2010-2014 Ingo Schwarze <schwarze@openbsd.org>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -28,17 +28,14 @@
 #include <string.h>
 
 #include "mandoc.h"
+#include "mandoc_aux.h"
 #include "out.h"
 #include "man.h"
 #include "term.h"
 #include "main.h"
 
-#define	INDENT		  7 /* fixed-width char full-indent */
-#define	HALFINDENT	  3 /* fixed-width char half-indent */
 #define	MAXMARGINS	  64 /* maximum number of indented scopes */
 
-/* FIXME: have PD set the default vspace width. */
-
 struct	mtermp {
 	int		  fl;
 #define	MANT_LITERAL	 (1 << 0)
@@ -46,12 +43,13 @@ struct	mtermp {
 	int		  lmargincur; /* index of current margin */
 	int		  lmarginsz; /* actual number of nested margins */
 	size_t		  offset; /* default offset to visible page */
+	int		  pardist; /* vert. space before par., unit: [v] */
 };
 
-#define	DECL_ARGS 	  struct termp *p, \
+#define	DECL_ARGS	  struct termp *p, \
 			  struct mtermp *mt, \
 			  const struct man_node *n, \
-			  const struct man_meta *m
+			  const struct man_meta *meta
 
 struct	termact {
 	int		(*pre)(DECL_ARGS);
@@ -67,24 +65,28 @@ static	void		  print_man_nodelist(DECL_ARGS);
 static	void		  print_man_node(DECL_ARGS);
 static	void		  print_man_head(struct termp *, const void *);
 static	void		  print_man_foot(struct termp *, const void *);
-static	void		  print_bvspace(struct termp *, 
-				const struct man_node *);
+static	void		  print_bvspace(struct termp *,
+				const struct man_node *, int);
 
-static	int		  pre_alternate(DECL_ARGS);
 static	int		  pre_B(DECL_ARGS);
 static	int		  pre_HP(DECL_ARGS);
 static	int		  pre_I(DECL_ARGS);
 static	int		  pre_IP(DECL_ARGS);
+static	int		  pre_OP(DECL_ARGS);
+static	int		  pre_PD(DECL_ARGS);
 static	int		  pre_PP(DECL_ARGS);
 static	int		  pre_RS(DECL_ARGS);
 static	int		  pre_SH(DECL_ARGS);
 static	int		  pre_SS(DECL_ARGS);
 static	int		  pre_TP(DECL_ARGS);
+static	int		  pre_UR(DECL_ARGS);
+static	int		  pre_alternate(DECL_ARGS);
+static	int		  pre_ft(DECL_ARGS);
 static	int		  pre_ign(DECL_ARGS);
 static	int		  pre_in(DECL_ARGS);
 static	int		  pre_literal(DECL_ARGS);
+static	int		  pre_ll(DECL_ARGS);
 static	int		  pre_sp(DECL_ARGS);
-static	int		  pre_ft(DECL_ARGS);
 
 static	void		  post_IP(DECL_ARGS);
 static	void		  post_HP(DECL_ARGS);
@@ -92,6 +94,7 @@ static	void		  post_RS(DECL_ARGS);
 static	void		  post_SH(DECL_ARGS);
 static	void		  post_SS(DECL_ARGS);
 static	void		  post_TP(DECL_ARGS);
+static	void		  post_UR(DECL_ARGS);
 
 static	const struct termact termacts[MAN_MAX] = {
 	{ pre_sp, NULL, MAN_NOTEXT }, /* br */
@@ -103,7 +106,7 @@ static	const struct termact termacts[MAN_MAX] = {
 	{ pre_PP, NULL, 0 }, /* PP */
 	{ pre_PP, NULL, 0 }, /* P */
 	{ pre_IP, post_IP, 0 }, /* IP */
-	{ pre_HP, post_HP, 0 }, /* HP */ 
+	{ pre_HP, post_HP, 0 }, /* HP */
 	{ NULL, NULL, 0 }, /* SM */
 	{ pre_B, NULL, 0 }, /* SB */
 	{ pre_alternate, NULL, 0 }, /* BI */
@@ -123,24 +126,32 @@ static	const struct termact termacts[MAN_MAX] = {
 	{ pre_RS, post_RS, 0 }, /* RS */
 	{ pre_ign, NULL, 0 }, /* DT */
 	{ pre_ign, NULL, 0 }, /* UC */
-	{ pre_ign, NULL, 0 }, /* PD */
+	{ pre_PD, NULL, MAN_NOTEXT }, /* PD */
 	{ pre_ign, NULL, 0 }, /* AT */
 	{ pre_in, NULL, MAN_NOTEXT }, /* in */
 	{ pre_ft, NULL, MAN_NOTEXT }, /* ft */
+	{ pre_OP, NULL, 0 }, /* OP */
+	{ pre_literal, NULL, 0 }, /* EX */
+	{ pre_literal, NULL, 0 }, /* EE */
+	{ pre_UR, post_UR, 0 }, /* UR */
+	{ NULL, NULL, 0 }, /* UE */
+	{ pre_ll, NULL, MAN_NOTEXT }, /* ll */
 };
 
 
-
 void
 terminal_man(void *arg, const struct man *man)
 {
 	struct termp		*p;
 	const struct man_node	*n;
-	const struct man_meta	*m;
+	const struct man_meta	*meta;
 	struct mtermp		 mt;
 
 	p = (struct termp *)arg;
 
+	if (0 == p->defindent)
+		p->defindent = 7;
+
 	p->overstep = 0;
 	p->maxrmargin = p->defrmargin;
 	p->tabwidth = term_len(p, 5);
@@ -149,18 +160,19 @@ terminal_man(void *arg, const struct man *man)
 		p->symtab = mchars_alloc();
 
 	n = man_node(man);
-	m = man_meta(man);
+	meta = man_meta(man);
 
-	term_begin(p, print_man_head, print_man_foot, m);
+	term_begin(p, print_man_head, print_man_foot, meta);
 	p->flags |= TERMP_NOSPACE;
 
 	memset(&mt, 0, sizeof(struct mtermp));
 
-	mt.lmargin[mt.lmargincur] = term_len(p, INDENT);
-	mt.offset = term_len(p, INDENT);
+	mt.lmargin[mt.lmargincur] = term_len(p, p->defindent);
+	mt.offset = term_len(p, p->defindent);
+	mt.pardist = 1;
 
 	if (n->child)
-		print_man_nodelist(p, &mt, n->child, m);
+		print_man_nodelist(p, &mt, n->child, meta);
 
 	term_end(p);
 }
@@ -177,7 +189,6 @@ a2height(const struct termp *p, const char *cp)
 	return(term_vspan(p, &su));
 }
 
-
 static int
 a2width(const struct termp *p, const char *cp)
 {
@@ -198,8 +209,9 @@ a2width(const struct termp *p, const char *cp)
  * first, print it.
  */
 static void
-print_bvspace(struct termp *p, const struct man_node *n)
+print_bvspace(struct termp *p, const struct man_node *n, int pardist)
 {
+	int	 i;
 
 	term_newln(p);
 
@@ -211,10 +223,11 @@ print_bvspace(struct termp *p, const struct man_node *n)
 		if (NULL == n->prev)
 			return;
 
-	term_vspace(p);
+	for (i = 0; i < pardist; i++)
+		term_vspace(p);
 }
 
-/* ARGSUSED */
+
 static int
 pre_ign(DECL_ARGS)
 {
@@ -222,8 +235,14 @@ pre_ign(DECL_ARGS)
 	return(0);
 }
 
+static int
+pre_ll(DECL_ARGS)
+{
+
+	term_setwidth(p, n->nchild ? n->child->string : NULL);
+	return(0);
+}
 
-/* ARGSUSED */
 static int
 pre_I(DECL_ARGS)
 {
@@ -232,15 +251,13 @@ pre_I(DECL_ARGS)
 	return(1);
 }
 
-
-/* ARGSUSED */
 static int
 pre_literal(DECL_ARGS)
 {
 
 	term_newln(p);
 
-	if (MAN_nf == n->tok)
+	if (MAN_nf == n->tok || MAN_EX == n->tok)
 		mt->fl |= MANT_LITERAL;
 	else
 		mt->fl &= ~MANT_LITERAL;
@@ -253,14 +270,28 @@ pre_literal(DECL_ARGS)
 	if (MAN_HP == n->parent->tok && p->rmargin < p->maxrmargin) {
 		p->offset = p->rmargin;
 		p->rmargin = p->maxrmargin;
-		p->flags &= ~(TERMP_NOBREAK | TERMP_TWOSPACE);
+		p->trailspace = 0;
+		p->flags &= ~(TERMP_NOBREAK | TERMP_BRIND);
 		p->flags |= TERMP_NOSPACE;
 	}
 
 	return(0);
 }
 
-/* ARGSUSED */
+static int
+pre_PD(DECL_ARGS)
+{
+
+	n = n->child;
+	if (0 == n) {
+		mt->pardist = 1;
+		return(0);
+	}
+	assert(MAN_TEXT == n->type);
+	mt->pardist = atoi(n->string);
+	return(0);
+}
+
 static int
 pre_alternate(DECL_ARGS)
 {
@@ -269,27 +300,27 @@ pre_alternate(DECL_ARGS)
 	int			 savelit, i;
 
 	switch (n->tok) {
-	case (MAN_RB):
+	case MAN_RB:
 		font[0] = TERMFONT_NONE;
 		font[1] = TERMFONT_BOLD;
 		break;
-	case (MAN_RI):
+	case MAN_RI:
 		font[0] = TERMFONT_NONE;
 		font[1] = TERMFONT_UNDER;
 		break;
-	case (MAN_BR):
+	case MAN_BR:
 		font[0] = TERMFONT_BOLD;
 		font[1] = TERMFONT_NONE;
 		break;
-	case (MAN_BI):
+	case MAN_BI:
 		font[0] = TERMFONT_BOLD;
 		font[1] = TERMFONT_UNDER;
 		break;
-	case (MAN_IR):
+	case MAN_IR:
 		font[0] = TERMFONT_UNDER;
 		font[1] = TERMFONT_NONE;
 		break;
-	case (MAN_IB):
+	case MAN_IB:
 		font[0] = TERMFONT_UNDER;
 		font[1] = TERMFONT_BOLD;
 		break;
@@ -304,7 +335,7 @@ pre_alternate(DECL_ARGS)
 		term_fontrepl(p, font[i]);
 		if (savelit && NULL == nn->next)
 			mt->fl |= MANT_LITERAL;
-		print_man_node(p, mt, nn, m);
+		print_man_node(p, mt, nn, meta);
 		if (nn->next)
 			p->flags |= TERMP_NOSPACE;
 	}
@@ -312,7 +343,6 @@ pre_alternate(DECL_ARGS)
 	return(0);
 }
 
-/* ARGSUSED */
 static int
 pre_B(DECL_ARGS)
 {
@@ -321,7 +351,28 @@ pre_B(DECL_ARGS)
 	return(1);
 }
 
-/* ARGSUSED */
+static int
+pre_OP(DECL_ARGS)
+{
+
+	term_word(p, "[");
+	p->flags |= TERMP_NOSPACE;
+
+	if (NULL != (n = n->child)) {
+		term_fontrepl(p, TERMFONT_BOLD);
+		term_word(p, n->string);
+	}
+	if (NULL != n && NULL != n->next) {
+		term_fontrepl(p, TERMFONT_UNDER);
+		term_word(p, n->next->string);
+	}
+
+	term_fontrepl(p, TERMFONT_NONE);
+	p->flags |= TERMP_NOSPACE;
+	term_word(p, "]");
+	return(0);
+}
+
 static int
 pre_ft(DECL_ARGS)
 {
@@ -334,26 +385,26 @@ pre_ft(DECL_ARGS)
 
 	cp = n->child->string;
 	switch (*cp) {
-	case ('4'):
+	case '4':
 		/* FALLTHROUGH */
-	case ('3'):
+	case '3':
 		/* FALLTHROUGH */
-	case ('B'):
+	case 'B':
 		term_fontrepl(p, TERMFONT_BOLD);
 		break;
-	case ('2'):
+	case '2':
 		/* FALLTHROUGH */
-	case ('I'):
+	case 'I':
 		term_fontrepl(p, TERMFONT_UNDER);
 		break;
-	case ('P'):
+	case 'P':
 		term_fontlast(p);
 		break;
-	case ('1'):
+	case '1':
 		/* FALLTHROUGH */
-	case ('C'):
+	case 'C':
 		/* FALLTHROUGH */
-	case ('R'):
+	case 'R':
 		term_fontrepl(p, TERMFONT_NONE);
 		break;
 	default:
@@ -362,7 +413,6 @@ pre_ft(DECL_ARGS)
 	return(0);
 }
 
-/* ARGSUSED */
 static int
 pre_in(DECL_ARGS)
 {
@@ -396,7 +446,7 @@ pre_in(DECL_ARGS)
 		p->offset -= p->offset > v ? v : p->offset;
 	else if (less > 0)
 		p->offset += v;
-	else 
+	else
 		p->offset = v;
 
 	/* Don't let this creep beyond the right margin. */
@@ -407,39 +457,61 @@ pre_in(DECL_ARGS)
 	return(0);
 }
 
-
-/* ARGSUSED */
 static int
 pre_sp(DECL_ARGS)
 {
+	char		*s;
 	size_t		 i, len;
+	int		 neg;
 
 	if ((NULL == n->prev && n->parent)) {
-		if (MAN_SS == n->parent->tok)
-			return(0);
-		if (MAN_SH == n->parent->tok)
+		switch (n->parent->tok) {
+		case MAN_SH:
+			/* FALLTHROUGH */
+		case MAN_SS:
+			/* FALLTHROUGH */
+		case MAN_PP:
+			/* FALLTHROUGH */
+		case MAN_LP:
+			/* FALLTHROUGH */
+		case MAN_P:
+			/* FALLTHROUGH */
 			return(0);
+		default:
+			break;
+		}
 	}
 
+	neg = 0;
 	switch (n->tok) {
-	case (MAN_br):
+	case MAN_br:
 		len = 0;
 		break;
 	default:
-		len = n->child ? a2height(p, n->child->string) : 1;
+		if (NULL == n->child) {
+			len = 1;
+			break;
+		}
+		s = n->child->string;
+		if ('-' == *s) {
+			neg = 1;
+			s++;
+		}
+		len = a2height(p, s);
 		break;
 	}
 
 	if (0 == len)
 		term_newln(p);
-	for (i = 0; i < len; i++)
-		term_vspace(p);
+	else if (neg)
+		p->skipvsp += len;
+	else
+		for (i = 0; i < len; i++)
+			term_vspace(p);
 
 	return(0);
 }
 
-
-/* ARGSUSED */
 static int
 pre_HP(DECL_ARGS)
 {
@@ -448,17 +520,20 @@ pre_HP(DECL_ARGS)
 	const struct man_node	*nn;
 
 	switch (n->type) {
-	case (MAN_BLOCK):
-		print_bvspace(p, n);
+	case MAN_BLOCK:
+		print_bvspace(p, n, mt->pardist);
 		return(1);
-	case (MAN_BODY):
-		p->flags |= TERMP_NOBREAK;
-		p->flags |= TERMP_TWOSPACE;
+	case MAN_BODY:
 		break;
 	default:
 		return(0);
 	}
 
+	if ( ! (MANT_LITERAL & mt->fl)) {
+		p->flags |= TERMP_NOBREAK | TERMP_BRIND;
+		p->trailspace = 2;
+	}
+
 	len = mt->lmargin[mt->lmargincur];
 	ival = -1;
 
@@ -481,20 +556,15 @@ pre_HP(DECL_ARGS)
 	return(1);
 }
 
-
-/* ARGSUSED */
 static void
 post_HP(DECL_ARGS)
 {
 
 	switch (n->type) {
-	case (MAN_BLOCK):
-		term_flushln(p);
-		break;
-	case (MAN_BODY):
-		term_flushln(p);
-		p->flags &= ~TERMP_NOBREAK;
-		p->flags &= ~TERMP_TWOSPACE;
+	case MAN_BODY:
+		term_newln(p);
+		p->flags &= ~(TERMP_NOBREAK | TERMP_BRIND);
+		p->trailspace = 0;
 		p->offset = mt->offset;
 		p->rmargin = p->maxrmargin;
 		break;
@@ -503,16 +573,14 @@ post_HP(DECL_ARGS)
 	}
 }
 
-
-/* ARGSUSED */
 static int
 pre_PP(DECL_ARGS)
 {
 
 	switch (n->type) {
-	case (MAN_BLOCK):
-		mt->lmargin[mt->lmargincur] = term_len(p, INDENT);
-		print_bvspace(p, n);
+	case MAN_BLOCK:
+		mt->lmargin[mt->lmargincur] = term_len(p, p->defindent);
+		print_bvspace(p, n, mt->pardist);
 		break;
 	default:
 		p->offset = mt->offset;
@@ -522,8 +590,6 @@ pre_PP(DECL_ARGS)
 	return(MAN_HEAD != n->type);
 }
 
-
-/* ARGSUSED */
 static int
 pre_IP(DECL_ARGS)
 {
@@ -532,14 +598,15 @@ pre_IP(DECL_ARGS)
 	int			 savelit, ival;
 
 	switch (n->type) {
-	case (MAN_BODY):
+	case MAN_BODY:
 		p->flags |= TERMP_NOSPACE;
 		break;
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		p->flags |= TERMP_NOBREAK;
+		p->trailspace = 1;
 		break;
-	case (MAN_BLOCK):
-		print_bvspace(p, n);
+	case MAN_BLOCK:
+		print_bvspace(p, n, mt->pardist);
 		/* FALLTHROUGH */
 	default:
 		return(1);
@@ -555,7 +622,7 @@ pre_IP(DECL_ARGS)
 				len = (size_t)ival;
 
 	switch (n->type) {
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		/* Handle zero-width lengths. */
 		if (0 == len)
 			len = term_len(p, 1);
@@ -572,15 +639,16 @@ pre_IP(DECL_ARGS)
 		mt->fl &= ~MANT_LITERAL;
 
 		if (n->child)
-			print_man_node(p, mt, n->child, m);
+			print_man_node(p, mt, n->child, meta);
 
 		if (savelit)
 			mt->fl |= MANT_LITERAL;
 
 		return(0);
-	case (MAN_BODY):
+	case MAN_BODY:
 		p->offset = mt->offset + len;
-		p->rmargin = p->maxrmargin;
+		p->rmargin = p->maxrmargin > p->offset ?
+		    p->maxrmargin : p->offset;
 		break;
 	default:
 		break;
@@ -589,28 +657,26 @@ pre_IP(DECL_ARGS)
 	return(1);
 }
 
-
-/* ARGSUSED */
 static void
 post_IP(DECL_ARGS)
 {
 
 	switch (n->type) {
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		term_flushln(p);
 		p->flags &= ~TERMP_NOBREAK;
+		p->trailspace = 0;
 		p->rmargin = p->maxrmargin;
 		break;
-	case (MAN_BODY):
+	case MAN_BODY:
 		term_newln(p);
+		p->offset = mt->offset;
 		break;
 	default:
 		break;
 	}
 }
 
-
-/* ARGSUSED */
 static int
 pre_TP(DECL_ARGS)
 {
@@ -619,14 +685,15 @@ pre_TP(DECL_ARGS)
 	int			 savelit, ival;
 
 	switch (n->type) {
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		p->flags |= TERMP_NOBREAK;
+		p->trailspace = 1;
 		break;
-	case (MAN_BODY):
+	case MAN_BODY:
 		p->flags |= TERMP_NOSPACE;
 		break;
-	case (MAN_BLOCK):
-		print_bvspace(p, n);
+	case MAN_BLOCK:
+		print_bvspace(p, n, mt->pardist);
 		/* FALLTHROUGH */
 	default:
 		return(1);
@@ -638,12 +705,12 @@ pre_TP(DECL_ARGS)
 	/* Calculate offset. */
 
 	if (NULL != (nn = n->parent->head->child))
-		if (nn->string && nn->parent->line == nn->line)
+		if (nn->string && 0 == (MAN_LINE & nn->flags))
 			if ((ival = a2width(p, nn->string)) >= 0)
 				len = (size_t)ival;
 
 	switch (n->type) {
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		/* Handle zero-length properly. */
 		if (0 == len)
 			len = term_len(p, 1);
@@ -655,9 +722,14 @@ pre_TP(DECL_ARGS)
 		mt->fl &= ~MANT_LITERAL;
 
 		/* Don't print same-line elements. */
-		for (nn = n->child; nn; nn = nn->next)
-			if (nn->line > n->line)
-				print_man_node(p, mt, nn, m);
+		nn = n->child;
+		while (NULL != nn && 0 == (MAN_LINE & nn->flags))
+			nn = nn->next;
+
+		while (NULL != nn) {
+			print_man_node(p, mt, nn, meta);
+			nn = nn->next;
+		}
 
 		if (savelit)
 			mt->fl |= MANT_LITERAL;
@@ -665,9 +737,12 @@ pre_TP(DECL_ARGS)
 			mt->lmargin[mt->lmargincur] = (size_t)ival;
 
 		return(0);
-	case (MAN_BODY):
+	case MAN_BODY:
 		p->offset = mt->offset + len;
-		p->rmargin = p->maxrmargin;
+		p->rmargin = p->maxrmargin > p->offset ?
+		    p->maxrmargin : p->offset;
+		p->trailspace = 0;
+		p->flags &= ~TERMP_NOBREAK;
 		break;
 	default:
 		break;
@@ -676,51 +751,47 @@ pre_TP(DECL_ARGS)
 	return(1);
 }
 
-
-/* ARGSUSED */
 static void
 post_TP(DECL_ARGS)
 {
 
 	switch (n->type) {
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		term_flushln(p);
-		p->flags &= ~TERMP_NOBREAK;
-		p->flags &= ~TERMP_TWOSPACE;
-		p->rmargin = p->maxrmargin;
 		break;
-	case (MAN_BODY):
+	case MAN_BODY:
 		term_newln(p);
+		p->offset = mt->offset;
 		break;
 	default:
 		break;
 	}
 }
 
-
-/* ARGSUSED */
 static int
 pre_SS(DECL_ARGS)
 {
+	int	 i;
 
 	switch (n->type) {
-	case (MAN_BLOCK):
+	case MAN_BLOCK:
 		mt->fl &= ~MANT_LITERAL;
-		mt->lmargin[mt->lmargincur] = term_len(p, INDENT);
-		mt->offset = term_len(p, INDENT);
+		mt->lmargin[mt->lmargincur] = term_len(p, p->defindent);
+		mt->offset = term_len(p, p->defindent);
 		/* If following a prior empty `SS', no vspace. */
 		if (n->prev && MAN_SS == n->prev->tok)
 			if (NULL == n->prev->body->child)
 				break;
 		if (NULL == n->prev)
 			break;
-		term_vspace(p);
+		for (i = 0; i < mt->pardist; i++)
+			term_vspace(p);
 		break;
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		term_fontrepl(p, TERMFONT_BOLD);
-		p->offset = term_len(p, HALFINDENT);
+		p->offset = term_len(p, 3);
 		break;
-	case (MAN_BODY):
+	case MAN_BODY:
 		p->offset = mt->offset;
 		break;
 	default:
@@ -730,17 +801,15 @@ pre_SS(DECL_ARGS)
 	return(1);
 }
 
-
-/* ARGSUSED */
 static void
 post_SS(DECL_ARGS)
 {
-	
+
 	switch (n->type) {
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		term_newln(p);
 		break;
-	case (MAN_BODY):
+	case MAN_BODY:
 		term_newln(p);
 		break;
 	default:
@@ -748,17 +817,16 @@ post_SS(DECL_ARGS)
 	}
 }
 
-
-/* ARGSUSED */
 static int
 pre_SH(DECL_ARGS)
 {
+	int	 i;
 
 	switch (n->type) {
-	case (MAN_BLOCK):
+	case MAN_BLOCK:
 		mt->fl &= ~MANT_LITERAL;
-		mt->lmargin[mt->lmargincur] = term_len(p, INDENT);
-		mt->offset = term_len(p, INDENT);
+		mt->lmargin[mt->lmargincur] = term_len(p, p->defindent);
+		mt->offset = term_len(p, p->defindent);
 		/* If following a prior empty `SH', no vspace. */
 		if (n->prev && MAN_SH == n->prev->tok)
 			if (NULL == n->prev->body->child)
@@ -766,13 +834,14 @@ pre_SH(DECL_ARGS)
 		/* If the first macro, no vspae. */
 		if (NULL == n->prev)
 			break;
-		term_vspace(p);
+		for (i = 0; i < mt->pardist; i++)
+			term_vspace(p);
 		break;
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		term_fontrepl(p, TERMFONT_BOLD);
 		p->offset = 0;
 		break;
-	case (MAN_BODY):
+	case MAN_BODY:
 		p->offset = mt->offset;
 		break;
 	default:
@@ -782,17 +851,15 @@ pre_SH(DECL_ARGS)
 	return(1);
 }
 
-
-/* ARGSUSED */
 static void
 post_SH(DECL_ARGS)
 {
-	
+
 	switch (n->type) {
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		term_newln(p);
 		break;
-	case (MAN_BODY):
+	case MAN_BODY:
 		term_newln(p);
 		break;
 	default:
@@ -800,7 +867,6 @@ post_SH(DECL_ARGS)
 	}
 }
 
-/* ARGSUSED */
 static int
 pre_RS(DECL_ARGS)
 {
@@ -808,24 +874,25 @@ pre_RS(DECL_ARGS)
 	size_t		 sz;
 
 	switch (n->type) {
-	case (MAN_BLOCK):
+	case MAN_BLOCK:
 		term_newln(p);
 		return(1);
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		return(0);
 	default:
 		break;
 	}
 
-	sz = term_len(p, INDENT);
+	sz = term_len(p, p->defindent);
 
 	if (NULL != (n = n->parent->head->child))
-		if ((ival = a2width(p, n->string)) >= 0) 
+		if ((ival = a2width(p, n->string)) >= 0)
 			sz = (size_t)ival;
 
 	mt->offset += sz;
-	p->rmargin = p->maxrmargin;
-	p->offset = mt->offset < p->rmargin ? mt->offset : p->rmargin;
+	p->offset = mt->offset;
+	p->rmargin = p->maxrmargin > p->offset ?
+	    p->maxrmargin : p->offset;
 
 	if (++mt->lmarginsz < MAXMARGINS)
 		mt->lmargincur = mt->lmarginsz;
@@ -834,7 +901,6 @@ pre_RS(DECL_ARGS)
 	return(1);
 }
 
-/* ARGSUSED */
 static void
 post_RS(DECL_ARGS)
 {
@@ -842,19 +908,19 @@ post_RS(DECL_ARGS)
 	size_t		 sz;
 
 	switch (n->type) {
-	case (MAN_BLOCK):
+	case MAN_BLOCK:
 		return;
-	case (MAN_HEAD):
+	case MAN_HEAD:
 		return;
 	default:
 		term_newln(p);
 		break;
 	}
 
-	sz = term_len(p, INDENT);
+	sz = term_len(p, p->defindent);
 
-	if (NULL != (n = n->parent->head->child)) 
-		if ((ival = a2width(p, n->string)) >= 0) 
+	if (NULL != (n = n->parent->head->child))
+		if ((ival = a2width(p, n->string)) >= 0)
 			sz = (size_t)ival;
 
 	mt->offset = mt->offset < sz ?  0 : mt->offset - sz;
@@ -864,6 +930,30 @@ post_RS(DECL_ARGS)
 		mt->lmargincur = mt->lmarginsz;
 }
 
+static int
+pre_UR(DECL_ARGS)
+{
+
+	return (MAN_HEAD != n->type);
+}
+
+static void
+post_UR(DECL_ARGS)
+{
+
+	if (MAN_BLOCK != n->type)
+		return;
+
+	term_word(p, "<");
+	p->flags |= TERMP_NOSPACE;
+
+	if (NULL != n->child->child)
+		print_man_node(p, mt, n->child->child, meta);
+
+	p->flags |= TERMP_NOSPACE;
+	term_word(p, ">");
+}
+
 static void
 print_man_node(DECL_ARGS)
 {
@@ -871,7 +961,7 @@ print_man_node(DECL_ARGS)
 	int		 c;
 
 	switch (n->type) {
-	case(MAN_TEXT):
+	case MAN_TEXT:
 		/*
 		 * If we have a blank line, output a vertical space.
 		 * If we have a space as the first character, break
@@ -884,38 +974,17 @@ print_man_node(DECL_ARGS)
 			term_newln(p);
 
 		term_word(p, n->string);
+		goto out;
 
-		/*
-		 * If we're in a literal context, make sure that words
-		 * togehter on the same line stay together.  This is a
-		 * POST-printing call, so we check the NEXT word.  Since
-		 * -man doesn't have nested macros, we don't need to be
-		 * more specific than this.
-		 */
-		if (MANT_LITERAL & mt->fl && ! (TERMP_NOBREAK & p->flags) &&
-				(NULL == n->next || 
-				 n->next->line > n->line)) {
-			rm = p->rmargin;
-			rmax = p->maxrmargin;
-			p->rmargin = p->maxrmargin = TERM_MAXMARGIN;
-			p->flags |= TERMP_NOSPACE;
-			term_flushln(p);
-			p->rmargin = rm;
-			p->maxrmargin = rmax;
-		}
-
-		if (MAN_EOS & n->flags)
-			p->flags |= TERMP_SENTENCE;
-		return;
-	case (MAN_EQN):
+	case MAN_EQN:
 		term_eqn(p, n->eqn);
 		return;
-	case (MAN_TBL):
+	case MAN_TBL:
 		/*
 		 * Tables are preceded by a newline.  Then process a
 		 * table line, which will cause line termination,
 		 */
-		if (TBL_SPAN_FIRST & n->span->flags) 
+		if (TBL_SPAN_FIRST & n->span->flags)
 			term_newln(p);
 		term_tbl(p, n->span);
 		return;
@@ -928,16 +997,41 @@ print_man_node(DECL_ARGS)
 
 	c = 1;
 	if (termacts[n->tok].pre)
-		c = (*termacts[n->tok].pre)(p, mt, n, m);
+		c = (*termacts[n->tok].pre)(p, mt, n, meta);
 
 	if (c && n->child)
-		print_man_nodelist(p, mt, n->child, m);
+		print_man_nodelist(p, mt, n->child, meta);
 
 	if (termacts[n->tok].post)
-		(*termacts[n->tok].post)(p, mt, n, m);
+		(*termacts[n->tok].post)(p, mt, n, meta);
 	if ( ! (MAN_NOTEXT & termacts[n->tok].flags))
 		term_fontrepl(p, TERMFONT_NONE);
 
+out:
+	/*
+	 * If we're in a literal context, make sure that words
+	 * together on the same line stay together.  This is a
+	 * POST-printing call, so we check the NEXT word.  Since
+	 * -man doesn't have nested macros, we don't need to be
+	 * more specific than this.
+	 */
+	if (MANT_LITERAL & mt->fl && ! (TERMP_NOBREAK & p->flags) &&
+	    (NULL == n->next || MAN_LINE & n->next->flags)) {
+		rm = p->rmargin;
+		rmax = p->maxrmargin;
+		p->rmargin = p->maxrmargin = TERM_MAXMARGIN;
+		p->flags |= TERMP_NOSPACE;
+		if (NULL != n->string && '\0' != *n->string)
+			term_flushln(p);
+		else
+			term_newln(p);
+		if (rm < rmax && n->parent->tok == MAN_HP) {
+			p->offset = rm;
+			p->rmargin = rmax;
+		} else
+			p->rmargin = rm;
+		p->maxrmargin = rmax;
+	}
 	if (MAN_EOS & n->flags)
 		p->flags |= TERMP_SENTENCE;
 }
@@ -947,96 +1041,128 @@ static void
 print_man_nodelist(DECL_ARGS)
 {
 
-	print_man_node(p, mt, n, m);
+	print_man_node(p, mt, n, meta);
 	if ( ! n->next)
 		return;
-	print_man_nodelist(p, mt, n->next, m);
+	print_man_nodelist(p, mt, n->next, meta);
 }
 
-
 static void
 print_man_foot(struct termp *p, const void *arg)
 {
-	const struct man_meta *meta;
+	const struct man_meta	*meta;
+	char			*title;
+	size_t			 datelen;
 
 	meta = (const struct man_meta *)arg;
+	assert(meta->title);
+	assert(meta->msec);
+	assert(meta->date);
 
 	term_fontrepl(p, TERMFONT_NONE);
 
-	term_vspace(p);
-	term_vspace(p);
-	term_vspace(p);
+	if (meta->hasbody)
+		term_vspace(p);
+
+	/*
+	 * Temporary, undocumented option to imitate mdoc(7) output.
+	 * In the bottom right corner, use the source instead of
+	 * the title.
+	 */
+
+	if ( ! p->mdocstyle) {
+		if (meta->hasbody) {
+			term_vspace(p);
+			term_vspace(p);
+		}
+		mandoc_asprintf(&title, "%s(%s)",
+		    meta->title, meta->msec);
+	} else if (meta->source) {
+		title = mandoc_strdup(meta->source);
+	} else {
+		title = mandoc_strdup("");
+	}
+	datelen = term_strlen(p, meta->date);
+
+	/* Bottom left corner: manual source. */
 
 	p->flags |= TERMP_NOSPACE | TERMP_NOBREAK;
-	p->rmargin = p->maxrmargin - term_strlen(p, meta->date);
+	p->trailspace = 1;
 	p->offset = 0;
-
-	/* term_strlen() can return zero. */
-	if (p->rmargin == p->maxrmargin)
-		p->rmargin--;
+	p->rmargin = (p->maxrmargin - datelen + term_len(p, 1)) / 2;
 
 	if (meta->source)
 		term_word(p, meta->source);
-	if (meta->source)
-		term_word(p, "");
 	term_flushln(p);
 
+	/* At the bottom in the middle: manual date. */
+
 	p->flags |= TERMP_NOSPACE;
 	p->offset = p->rmargin;
-	p->rmargin = p->maxrmargin;
-	p->flags &= ~TERMP_NOBREAK;
+	p->rmargin = p->maxrmargin - term_strlen(p, title);
+	if (p->offset + datelen >= p->rmargin)
+		p->rmargin = p->offset + datelen;
 
 	term_word(p, meta->date);
 	term_flushln(p);
-}
 
+	/* Bottom right corner: manual title and section. */
+
+	p->flags &= ~TERMP_NOBREAK;
+	p->flags |= TERMP_NOSPACE;
+	p->trailspace = 0;
+	p->offset = p->rmargin;
+	p->rmargin = p->maxrmargin;
+
+	term_word(p, title);
+	term_flushln(p);
+	free(title);
+}
 
 static void
 print_man_head(struct termp *p, const void *arg)
 {
-	char		buf[BUFSIZ], title[BUFSIZ];
-	size_t		buflen, titlen;
-	const struct man_meta *m;
+	const struct man_meta	*meta;
+	const char		*volume;
+	char			*title;
+	size_t			 vollen, titlen;
 
-	m = (const struct man_meta *)arg;
+	meta = (const struct man_meta *)arg;
+	assert(meta->title);
+	assert(meta->msec);
 
-	/*
-	 * Note that old groff would spit out some spaces before the
-	 * header.  We discontinue this strange behaviour, but at one
-	 * point we did so here.
-	 */
+	volume = NULL == meta->vol ? "" : meta->vol;
+	vollen = term_strlen(p, volume);
 
-	p->offset = 0;
-	p->rmargin = p->maxrmargin;
+	/* Top left corner: manual title and section. */
 
-	buf[0] = title[0] = '\0';
-
-	if (m->vol)
-		strlcpy(buf, m->vol, BUFSIZ);
-	buflen = term_strlen(p, buf);
-
-	snprintf(title, BUFSIZ, "%s(%s)", m->title, m->msec);
+	mandoc_asprintf(&title, "%s(%s)", meta->title, meta->msec);
 	titlen = term_strlen(p, title);
 
 	p->flags |= TERMP_NOBREAK | TERMP_NOSPACE;
+	p->trailspace = 1;
 	p->offset = 0;
-	p->rmargin = 2 * (titlen+1) + buflen < p->maxrmargin ?
-	    (p->maxrmargin - 
-	     term_strlen(p, buf) + term_len(p, 1)) / 2 :
-	    p->maxrmargin - buflen;
+	p->rmargin = 2 * (titlen+1) + vollen < p->maxrmargin ?
+	    (p->maxrmargin - vollen + term_len(p, 1)) / 2 :
+	    p->maxrmargin - vollen;
 
 	term_word(p, title);
 	term_flushln(p);
 
+	/* At the top in the middle: manual volume. */
+
 	p->flags |= TERMP_NOSPACE;
 	p->offset = p->rmargin;
-	p->rmargin = p->offset + buflen + titlen < p->maxrmargin ?
+	p->rmargin = p->offset + vollen + titlen < p->maxrmargin ?
 	    p->maxrmargin - titlen : p->maxrmargin;
 
-	term_word(p, buf);
+	term_word(p, volume);
 	term_flushln(p);
 
+	/* Top right corner: title and section, again. */
+
 	p->flags &= ~TERMP_NOBREAK;
+	p->trailspace = 0;
 	if (p->rmargin + titlen <= p->maxrmargin) {
 		p->flags |= TERMP_NOSPACE;
 		p->offset = p->rmargin;
@@ -1049,12 +1175,16 @@ print_man_head(struct termp *p, const void *arg)
 	p->offset = 0;
 	p->rmargin = p->maxrmargin;
 
-	/* 
-	 * Groff likes to have some leading spaces before content.  Well
-	 * that's fine by me.
+	/*
+	 * Groff prints three blank lines before the content.
+	 * Do the same, except in the temporary, undocumented
+	 * mode imitating mdoc(7) output.
 	 */
 
 	term_vspace(p);
-	term_vspace(p);
-	term_vspace(p);
+	if ( ! p->mdocstyle) {
+		term_vspace(p);
+		term_vspace(p);
+	}
+	free(title);
 }
