@@ -1,4 +1,4 @@
-/*	$Id: msec.c,v 1.9 2011/03/22 14:33:05 kristaps Exp $ */
+/*	$Id: msec.c,v 1.11 2014/03/23 11:25:26 schwarze Exp $ */
 /*
  * Copyright (c) 2009 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -18,18 +18,16 @@
 #include "config.h"
 #endif
 
-#include <stdlib.h>
 #include <string.h>
 
-#include "mdoc.h"
 #include "mandoc.h"
-#include "libmdoc.h"
+#include "libmandoc.h"
 
 #define LINE(x, y) \
 	if (0 == strcmp(p, x)) return(y);
 
 const char *
-mdoc_a2msec(const char *p)
+mandoc_a2msec(const char *p)
 {
 
 #include "msec.in"
